from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import NamedTuple
from typing import Optional

import typer

from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.exceptions import ZabbixAPICallError
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.output.console import error
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.render import render_result
from zabbix_cli.utils.args import get_hostgroup_hosts
from zabbix_cli.utils.args import parse_int_list_arg
from zabbix_cli.utils.args import parse_list_arg
from zabbix_cli.utils.utils import compile_pattern

if TYPE_CHECKING:
    from zabbix_cli.app import StatefulApp
    from zabbix_cli.pyzabbix.types import Host
    from zabbix_cli.pyzabbix.types import Proxy

HELP_PANEL = "Proxy"
HELP_PANEL_GROUP = "Proxy Group"


class PrevProxyHosts(NamedTuple):
    hosts: list[Host]
    proxy: Optional[Proxy] = None


def ensure_proxy_group_support() -> None:
    if app.state.client.version.major < 7:
        exit_err("Proxy groups require Zabbix 7.0 or later.")


def group_hosts_by_proxy(
    app: StatefulApp, hosts: list[Host], default_proxy_id: str = ""
) -> dict[str, PrevProxyHosts]:
    """Group hosts by the proxy they had prior to the update."""
    proxy_ids: set[str] = set()
    for host in hosts:
        if host.proxyid:
            proxy_ids.add(host.proxyid)

    # Fetch proxies for all observed proxy IDs
    proxy_mapping: dict[str, PrevProxyHosts] = {}
    for proxy_id in proxy_ids:
        try:
            p = app.state.client.get_proxy(proxy_id)
        except ZabbixAPICallError as e:
            # Should be nigh-impossible, but someone might delete the proxy
            # while the command is running
            error(f"{e}")
            p = None
        proxy_mapping[proxy_id] = PrevProxyHosts(hosts=[], proxy=p)
    # The default is a special case - no prev proxy exists for these hosts
    proxy_mapping[default_proxy_id] = PrevProxyHosts(hosts=[], proxy=None)

    for host in hosts:
        if not host.proxyid:
            host_proxyid = default_proxy_id
        else:
            host_proxyid = host.proxyid
        proxy_mapping[host_proxyid].hosts.append(host)

    # No hosts without previous proxy - remove from mapping
    if not proxy_mapping[default_proxy_id].hosts:
        del proxy_mapping[default_proxy_id]

    return proxy_mapping


@app.command(name="clear_host_proxy", rich_help_panel=HELP_PANEL)
def clear_host_proxy(
    ctx: typer.Context,
    hostname: str = typer.Argument(
        help="Hostnames. Comma-separated Supports wildcards.",
        show_default=False,
    ),
    dryrun: bool = typer.Option(False, help="Preview changes"),
) -> None:
    """Clear the proxy for one or more hosts.

    Sets the hosts to be monitored by the Zabbix server instead of a proxy.
    """
    # NOTE: this command is _VERY_ similar to `update_host_proxy`
    #       can we refactor them to avoid code duplication?
    from zabbix_cli.commands.results.proxy import ClearHostProxyResult
    from zabbix_cli.models import AggregateResult
    from zabbix_cli.output.console import error
    from zabbix_cli.output.console import info

    hostnames = parse_list_arg(hostname)
    hosts = app.state.client.get_hosts(*hostnames, search=True)

    to_update = [host for host in hosts if host.proxyid]
    if not dryrun:
        with app.status("Clearing host proxies..."):
            if not to_update:
                exit_err("No matching hosts have a proxy assigned.")
            try:
                app.state.client.clear_host_proxies(to_update)
            except Exception as e:
                error(f"Failed to clear proxies for hosts: {e}")

    proxy_map = group_hosts_by_proxy(app, to_update)

    render_result(
        AggregateResult(
            empty_ok=True,
            result=[
                ClearHostProxyResult.from_result(
                    hosts=prev.hosts,
                    source_proxy=prev.proxy,
                )
                for _, prev in proxy_map.items()
            ],
        )
    )

    total_hosts = len(hosts)
    if dryrun:
        info(f"Would clear proxy for {total_hosts} hosts.")
    else:
        success(f"Cleared proxy for {total_hosts} hosts.")


@app.command(
    name="load_balance_proxy_hosts",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Load balance hosts evenly between two proxies",
            "load_balance_proxy_hosts proxy1,proxy2",
        ),
        Example(
            "Place twice as many hosts on proxy1 as proxy2",
            "load_balance_proxy_hosts proxy1,proxy2 2,1",
        ),
        Example(
            "Load balance hosts evenly between three proxies",
            "load_balance_proxy_hosts proxy1,proxy2,proxy3",
        ),
        Example(
            "Load balance hosts unevenly between three proxies",
            "load_balance_proxy_hosts proxy1,proxy2,proxy3 1,1,2",
        ),
    ],
)
def load_balance_proxy_hosts(
    ctx: typer.Context,
    proxy: str = typer.Argument(
        help="Comma delimited list of proxies to share hosts between.",
        metavar="<proxy1,proxy2,...>",
        show_default=False,
    ),
    weight: Optional[str] = typer.Argument(
        None,
        help="Weights for each proxy. Comma-separated. Defaults to equal weights.",
        metavar="[weight1,weight2,...]",
        show_default=False,
    ),
) -> None:
    """Spread hosts between multiple proxies.

    Hosts are determined by the hosts monitored by the given proxies
    Hosts monitored by other proxies or not monitored at all are not affected.

    Weighting for the load balancing is optional, and defaults to equal weights.
    Number of proxies must match number of weights if specified.
    """
    import itertools
    import random

    from zabbix_cli.commands.results.proxy import LBProxy
    from zabbix_cli.commands.results.proxy import LBProxyResult

    proxy_names = [p.strip() for p in proxy.split(",")]
    if weight:
        weights = parse_int_list_arg(weight)
    else:
        weights = [1] * len(proxy_names)  # default to equal weights

    if len(proxy_names) != len(weights):
        exit_err("Number of proxies must match number of weights.")
    elif len(proxy_names) < 2:
        exit_err("Must specify at least two proxies to load balance.")
    elif all(w == 0 for w in weights):
        exit_err("All weights cannot be zero.")

    # Fetch proxies one by one to ensure each one exists
    proxies = [app.state.client.get_proxy(p, select_hosts=True) for p in proxy_names]

    # TODO: Make sure list of proxies is in same order we specified them?

    all_hosts = list(itertools.chain.from_iterable(p.hosts for p in proxies))
    if not all_hosts:
        exit_err("Proxies have no hosts to load balance.")
    logging.debug("Found %d hosts to load balance.", len(all_hosts))

    lb_proxies = {
        p.proxyid: LBProxy(proxy=p, weight=w) for p, w in zip(proxies, weights)
    }
    for host in all_hosts:
        p = random.choices(proxies, weights=weights, k=1)[0]
        lb_proxies[p.proxyid].hosts.append(host)

    # Abort on failure
    try:
        for lb_proxy in lb_proxies.values():
            n_hosts = len(lb_proxy.hosts)
            logging.debug(
                "Proxy '%s' has %d hosts after balancing.",
                lb_proxy.proxy.name,
                n_hosts,
            )
            if not n_hosts:
                logging.debug(
                    "Proxy '%s' has no hosts after balancing.", lb_proxy.proxy.name
                )
                continue
            logging.debug("Moving %d hosts to proxy %r", n_hosts, lb_proxy.proxy.name)

            app.state.client.move_hosts_to_proxy(
                hosts=lb_proxy.hosts,
                proxy=lb_proxy.proxy,
            )
            lb_proxy.count = n_hosts
    except Exception as e:
        raise ZabbixCLIError(f"Failed to load balance hosts: {e}") from e

    render_result(LBProxyResult(proxies=list(lb_proxies.values())))
    # HACK: render_result doesn't print a message for table results
    success(f"Load balanced {len(all_hosts)} hosts between {len(proxies)} proxies.")


@app.command(
    name="move_proxy_hosts",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Move all hosts from one proxy to another",
            "move_proxy_hosts proxy1 proxy2",
        ),
        Example(
            "Move all hosts with names matching a regex pattern",
            "move_proxy_hosts proxy1 proxy2 --filter '$www.*'",
        ),
    ],
)
def move_proxy_hosts(
    ctx: typer.Context,
    proxy_src: str = typer.Argument(
        help="Proxy to move hosts from.",
        show_default=False,
    ),
    proxy_dst: str = typer.Argument(
        help="Proxy to move hosts to.",
        show_default=False,
    ),
    # Prefer --filter over positional arg
    host_filter: Optional[str] = typer.Option(
        None, "--filter", help="Regex pattern of hosts to move."
    ),
    # LEGACY: matches old command signature (deprecated)
    host_filter_arg: Optional[str] = typer.Argument(
        None, help="Filter hosts to move.", hidden=True
    ),
) -> None:
    """Move hosts from one proxy to another."""
    from zabbix_cli.commands.results.proxy import MoveProxyHostsResult
    from zabbix_cli.models import Result

    hfilter = host_filter_arg or host_filter
    if hfilter:  # Compile before we fetch to fail fast
        filter_pattern = compile_pattern(hfilter)
    else:
        filter_pattern = None

    source_proxy = app.state.client.get_proxy(proxy_src)
    destination_proxy = app.state.client.get_proxy(proxy_dst)

    hosts = app.state.client.get_hosts(proxy=source_proxy)
    if not hosts:
        exit_err(f"Source proxy {source_proxy.name!r} has no hosts.")

    # Do filtering client-side to get full regex support
    if filter_pattern:
        hosts = [host for host in hosts if filter_pattern.match(host.host)]
        if not hosts:
            exit_err(f"No hosts matched filter {hfilter!r}.")

    app.state.client.move_hosts_to_proxy(hosts, destination_proxy)

    render_result(
        Result(
            message=f"Moved {len(hosts)} hosts from {source_proxy.name!r} to {destination_proxy.name!r}",
            result=MoveProxyHostsResult(
                source=source_proxy.proxyid,
                destination=destination_proxy.proxyid,
                hosts=[host.host for host in hosts],
            ),
        )
    )


@app.command(name="show_proxies", rich_help_panel=HELP_PANEL)
def show_proxies(
    ctx: typer.Context,
    name_or_id: Optional[str] = typer.Argument(
        None,
        help="Filter by proxy name or ID. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    hosts: bool = typer.Option(
        False,
        "--hosts",
        help="Show hostnames of each host for every proxy.",
    ),
) -> None:
    """Show all proxies.

    Shows number of hosts for each proxy unless [option]--hosts[/] is passed in,
    in which case the hostnames of each host are displayed instead.
    """
    from zabbix_cli.commands.results.proxy import ShowProxiesResult
    from zabbix_cli.models import AggregateResult

    names_or_ids = parse_list_arg(name_or_id)

    with app.status("Fetching proxies..."):
        proxies = app.state.client.get_proxies(
            *names_or_ids,
            select_hosts=True,
        )
    render_result(
        AggregateResult(
            result=[ShowProxiesResult.from_result(p, show_hosts=hosts) for p in proxies]
        )
    )


@app.command(name="show_proxy_hosts", rich_help_panel=HELP_PANEL)
def show_proxy_hosts(
    ctx: typer.Context,
    proxy: str = typer.Argument(
        help="Proxy name or ID. Supports wildcards.",
        show_default=False,
    ),
) -> None:
    """Show all hosts with for a given proxy."""
    from zabbix_cli.commands.results.proxy import ShowProxyHostsResult

    with app.status("Fetching proxy..."):
        prox = app.state.client.get_proxy(proxy, select_hosts=True)

    render_result(ShowProxyHostsResult.from_result(prox))


@app.command(name="update_host_proxy", rich_help_panel=HELP_PANEL)
def update_host_proxy(
    ctx: typer.Context,
    hostname: str = typer.Argument(
        help="Hostnames. Comma-separated Supports wildcards.",
        show_default=False,
    ),
    proxy: str = typer.Argument(
        help="Proxy name. Supports wildcards.",
        show_default=False,
    ),
    dryrun: bool = typer.Option(
        False,
        help="Preview changes",
    ),
) -> None:
    """Assign hosts to a proxy.

    Supports wildcards for both hosts and proxy names.
    If multiple proxies match the proxy name, the first match is used.
    """
    from zabbix_cli.commands.results.proxy import UpdateHostProxyResult
    from zabbix_cli.models import AggregateResult
    from zabbix_cli.output.console import info

    hostnames = parse_list_arg(hostname)
    hosts = app.state.client.get_hosts(*hostnames, search=True)
    dest_proxy = app.state.client.get_proxy(proxy)

    to_update: list[Host] = []
    for host in hosts:
        if host.proxyid != dest_proxy.proxyid:
            to_update.append(host)

    if not dryrun:
        with app.status("Updating host proxies..."):
            hostids = app.state.client.update_hosts_proxy(to_update, dest_proxy)
    else:
        hostids = [host.hostid for host in to_update]

    updated = set(hostids)
    updated_hosts = [host for host in to_update if host.hostid in updated]

    proxy_hosts = group_hosts_by_proxy(app, updated_hosts)

    render_result(
        AggregateResult(
            empty_ok=True,
            result=[
                UpdateHostProxyResult.from_result(
                    hosts=prev.hosts,
                    source_proxy=prev.proxy,
                    dest_proxy=dest_proxy,
                )
                for _, prev in proxy_hosts.items()
            ],
        )
    )

    total_hosts = len(updated)
    if dryrun:
        info(f"Would update proxy for {total_hosts} hosts.")
    else:
        success(f"Updated proxy for {total_hosts} hosts.")


@app.command(name="update_hostgroup_proxy", rich_help_panel=HELP_PANEL)
def update_hostgroup_proxy(
    ctx: typer.Context,
    hostgroup: str = typer.Argument(
        help="Host group(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    proxy: str = typer.Argument(
        help="Proxy to assign. Supports wildcards.",
        show_default=False,
    ),
    dryrun: bool = typer.Option(False, help="Preview changes."),
) -> None:
    """Assign a proxy to all hosts in one or more host groups."""
    from zabbix_cli.commands.results.proxy import UpdateHostGroupProxyResult
    from zabbix_cli.output.console import warning

    prx = app.state.client.get_proxy(proxy)

    hosts = get_hostgroup_hosts(app, hostgroup)
    to_update: list[Host] = []
    for host in hosts:
        if host.proxyid != prx.proxyid:
            to_update.append(host)

    if not dryrun:
        if not to_update:
            exit_err("All hosts already have the specified proxy.")
        with app.status("Updating host proxies..."):
            hostids = app.state.client.update_hosts_proxy(to_update, prx)
        if not hostids:
            warning("No hosts were updated.")
    else:
        hostids = [host.hostid for host in to_update]

    updated_hosts = [host for host in to_update if host.hostid in hostids]
    render_result(UpdateHostGroupProxyResult.from_result(prx, updated_hosts))

    total_hosts = len(hostids)
    if dryrun:
        info(f"Would update proxy for {total_hosts} hosts.")
    else:
        success(f"Updated proxy for {total_hosts} hosts.")


@app.command(
    name="add_proxy_to_group",
    rich_help_panel=HELP_PANEL_GROUP,
    examples=[
        Example(
            "Add a proxy to a proxy group",
            "add_proxy_to_group proxy1 group1 192.168.0.2 10051",
        )
    ],
)
def add_proxy_to_group(
    ctx: typer.Context,
    name_or_id: str = typer.Argument(
        help="Name or ID of proxy to add.",
        show_default=False,
    ),
    proxy_group: str = typer.Argument(
        help="Name or ID of proxy group to add proxy to.",
        show_default=False,
    ),
    local_address: Optional[str] = typer.Argument(
        None,
        help="Address for active agents.",
        show_default=False,
    ),
    local_port: Optional[str] = typer.Argument(
        None,
        help="Address for active agents.",
        show_default=False,
    ),
) -> None:
    """Add a proxy to a proxy group.

    Requires a local address and port for active agent redirection if
    if the proxy does not have it set.
    """
    ensure_proxy_group_support()

    proxy = app.state.client.get_proxy(name_or_id)

    # Determine address + port
    local_address = local_address or proxy.local_address
    if not local_address:
        exit_err(f"Proxy {proxy.name} requires a local address for active agents.")
    local_port = local_port or proxy.local_port
    if not local_port:
        exit_err(f"Proxy {proxy.name} requires a local port for active agents.")

    group = app.state.client.get_proxy_group(proxy_group)
    app.state.client.add_proxy_to_group(proxy, group, local_address, local_port)

    success(f"Added proxy {proxy.name!r} to group {group.name!r}.")


@app.command(name="remove_proxy_from_group", rich_help_panel=HELP_PANEL_GROUP)
def remove_proxy_from_group(
    ctx: typer.Context,
    name_or_id: str = typer.Argument(
        help="Name or ID of proxy to remove.",
        show_default=False,
    ),
) -> None:
    """Remove a proxy from a proxy group."""
    ensure_proxy_group_support()

    proxy = app.state.client.get_proxy(name_or_id)
    if proxy.proxy_groupid is None or proxy.proxy_groupid == "0":
        exit_err(f"Proxy {proxy.name!r} is not in a proxy group.")

    app.state.client.remove_proxy_from_group(proxy)

    success(f"Removed proxy {proxy.name!r} from group with ID {proxy.proxy_groupid}.")


@app.command(
    name="show_proxy_groups",
    rich_help_panel=HELP_PANEL_GROUP,
    examples=[
        Example(
            "Show all proxy groups",
            "show_proxy_groups",
        ),
        Example(
            "Show proxy groups with a specific proxy",
            "show_proxy_groups --proxy proxy1",
        ),
        Example(
            "Show proxy groups with either proxy1 or proxy2",
            "show_proxy_groups --proxy proxy1,proxy2",
        ),
    ],
)
def show_proxy_groups(
    ctx: typer.Context,
    name_or_id: Optional[str] = typer.Argument(
        None,
        help="Filter by proxy name or ID. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    proxies: Optional[str] = typer.Option(
        None,
        "--proxy",
        help="Show only groups containing these proxies. Comma-separated.",
    ),
) -> None:
    """Show all proxy groups and their proxies.

    Optionally takes in a list of names or IDs to filter by.
    """
    from zabbix_cli.models import AggregateResult

    ensure_proxy_group_support()

    names_or_ids = parse_list_arg(name_or_id)
    proxy_names = parse_list_arg(proxies)

    with app.status("Fetching proxy groups..."):
        # Fetch proxies if specified
        proxy_list = [app.state.client.get_proxy(p) for p in proxy_names]
        groups = app.state.client.get_proxy_groups(
            *names_or_ids,
            proxies=proxy_list,
            select_proxies=True,
        )

    render_result(AggregateResult(result=groups))


@app.command(name="show_proxy_group_hosts", rich_help_panel=HELP_PANEL_GROUP)
def show_proxy_group_hosts(
    ctx: typer.Context,
    proxygroup: str = typer.Argument(
        help="Proxy group name or ID. Supports wildcards.",
        show_default=False,
    ),
) -> None:
    """Show all hosts in a proxy group."""
    from zabbix_cli.commands.results.proxy import ShowProxyGroupHostsResult

    ensure_proxy_group_support()

    with app.status("Fetching proxy groups...") as status:
        group = app.state.client.get_proxy_group(proxygroup)
        status.update("Fetching hosts...")
        hosts = app.state.client.get_hosts(proxy_group=group)

    render_result(ShowProxyGroupHostsResult(proxy_group=group, hosts=hosts))


@app.command(name="update_hostgroup_proxygroup", rich_help_panel=HELP_PANEL_GROUP)
def update_hostgroup_proxygroup(
    ctx: typer.Context,
    hostgroup: str = typer.Argument(
        help="Host group(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    proxygroup: str = typer.Argument(
        help="Proxy group to assign. Supports wildcards.",
        show_default=False,
    ),
    dryrun: bool = typer.Option(False, help="Preview changes."),
) -> None:
    """Assign a proxy group to all hosts in one or more host groups."""
    from zabbix_cli.commands.results.proxy import UpdateHostGroupProxyGroupResult
    from zabbix_cli.output.console import warning

    ensure_proxy_group_support()

    grp = app.state.client.get_proxy_group(proxygroup)

    hosts = get_hostgroup_hosts(app, hostgroup)
    to_update: list[Host] = []
    for host in hosts:
        if host.proxy_groupid != grp.proxy_groupid:
            to_update.append(host)

    # Sort hosts by host group
    updated: list[str] = []  # list of host IDs
    if not dryrun:
        if not to_update:
            exit_err("All hosts already have the specified proxy group.")
        with app.status("Updating hosts..."):
            updated = app.state.client.add_hosts_to_proxygroup(to_update, grp)
        if not updated:
            warning("No hosts were updated.")
    else:
        updated = [host.hostid for host in to_update]

    updated_hosts = [host for host in to_update if host.hostid in updated]
    render_result(UpdateHostGroupProxyGroupResult.from_result(grp, updated_hosts))

    total_hosts = len(updated)
    if dryrun:
        info(f"Would update proxy group for {total_hosts} hosts.")
    else:
        success(f"Updated proxy group for {total_hosts} hosts.")



"""Commands for managing user groups."""

from __future__ import annotations

import logging
from contextlib import suppress
from typing import TYPE_CHECKING
from typing import Optional
from typing import TypeVar

import typer
from strenum import StrEnum

from zabbix_cli._v2_compat import ARGS_POSITIONAL
from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.commands.common.args import OPTION_LIMIT
from zabbix_cli.exceptions import ZabbixAPIException
from zabbix_cli.exceptions import ZabbixNotFoundError
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.console import success
from zabbix_cli.output.console import warning
from zabbix_cli.output.formatting.grammar import pluralize as p
from zabbix_cli.output.render import render_result
from zabbix_cli.pyzabbix.enums import GUIAccess
from zabbix_cli.pyzabbix.enums import UsergroupPermission
from zabbix_cli.utils.args import parse_bool_arg
from zabbix_cli.utils.args import parse_list_arg

if TYPE_CHECKING:
    from typing import Protocol

    class UsergroupLike(Protocol):
        name: str
        usrgrpid: str

    UsergroupLikeT = TypeVar("UsergroupLikeT", bound=UsergroupLike)


class UsergroupSorting(StrEnum):
    NAME = "name"
    ID = "id"
    USERS = "users"


def sort_ugroups(
    ugroups: list[UsergroupLikeT], sort: UsergroupSorting
) -> list[UsergroupLikeT]:
    """Sort result types based on user group objects.

    I.e. we have some custom types that all share the samse attributes
    `name` and `usrgrpid`. This function sorts them based on the given
    sorting method.
    """
    if sort == UsergroupSorting.NAME:
        return sorted(ugroups, key=lambda ug: ug.name)
    elif sort == UsergroupSorting.ID:
        # NOTE: this can fail if user group IDs are not integers (API returns strings)
        # We should potentially just coerce them to ints in the model
        try:
            return sorted(ugroups, key=lambda ug: int(ug.usrgrpid))
        except Exception as e:
            logging.error("Failed to sort user groups by ID: %s", e)
            # Fall back on unsorted (likely sorted by ID anyway)
    return ugroups


OPTION_SORT_UGROUPS = typer.Option(
    UsergroupSorting.NAME,
    "--sort",
    help="Sort by field.",
    case_sensitive=False,
)

HELP_PANEL = "User Group"


@app.command("add_user_to_usergroup", rich_help_panel=HELP_PANEL)
def add_user_to_usergroup(
    ctx: typer.Context,
    usernames: str = typer.Argument(
        help="Usernames to add. Comma-separated.",
        show_default=False,
    ),
    usergroups: str = typer.Argument(
        help="User groups to add the users to. Comma-separated.",
        show_default=False,
    ),
) -> None:
    """Add users to usergroups.

    Users and groups must exist.
    """
    from zabbix_cli.commands.results.usergroup import UsergroupAddUsers

    # FIXME: requires support for IDs for parity with V2
    unames = parse_list_arg(usernames)
    ugroups = parse_list_arg(usergroups)

    with app.status("Adding users to user groups..."):
        users = [app.state.client.get_user(u) for u in unames]
        for ugroup in ugroups:
            try:
                app.state.client.add_usergroup_users(ugroup, users)
            except ZabbixAPIException as e:
                exit_err(f"Failed to add users to user group {ugroup!r}: {e}")

    render_result(
        UsergroupAddUsers(
            usergroups=ugroups,
            users=[u.username for u in users],
        ),
    )
    success("Added users to user groups.")


# NOTE: {add,update}_usergroup_permissions seem to be the exact same command in V2. Keeping that here.
@app.command("add_usergroup_permissions", rich_help_panel=HELP_PANEL)
@app.command("update_usergroup_permissions", rich_help_panel=HELP_PANEL, hidden=True)
def add_usergroup_permissions(
    ctx: typer.Context,
    usergroup: str = typer.Argument(
        help="User group to give permissions to.",
        show_default=False,
    ),
    hostgroups: Optional[str] = typer.Option(
        None,
        "--hostgroup",
        help="Comma-separated list of host group names.",
    ),
    templategroups: Optional[str] = typer.Option(
        None,
        "--templategroup",
        help="Comma-separated list of template group names.",
    ),
    permission: Optional[UsergroupPermission] = typer.Option(
        None,
        "--permission",
        help="Permission to give to the user group.",
        case_sensitive=False,
    ),
    # Legacy V2 args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Give a user group permissions to host/template groups.

    Run [command]show_hostgroups[/] to get a list of host groups, and
    [command]show_templategroups --no-templates[/] to get a list of template groups.
    """
    from zabbix_cli.commands.results.usergroup import AddUsergroupPermissionsResult

    # Legacy positional args: <usergroup> <hostgroups> <permission>
    # We already have usergroup as positional arg, so we are left with 2 args.
    if args:
        warning("Positional arguments are deprecated. Please use options instead.")
        if len(args) != 2:
            exit_err(
                "Invalid number of positional arguments. Please use options instead."
            )
        hostgroups = hostgroups or args[0]
        permission = permission or UsergroupPermission(args[1])

    hgroups = parse_list_arg(hostgroups)
    tgroups = parse_list_arg(templategroups)

    if not hgroups and not tgroups:
        exit_err("At least one host group or template group must be specified.")

    if not permission:
        permission = UsergroupPermission.from_prompt(
            default=UsergroupPermission.READ_WRITE
        )

    if hgroups:
        with app.status("Adding host group permissions..."):
            app.state.client.update_usergroup_rights(
                usergroup, hgroups, permission, hostgroup=True
            )
        success("Added host group permissions.")

    if tgroups:
        with app.status("Adding template group permissions..."):
            app.state.client.update_usergroup_rights(
                usergroup, tgroups, permission, hostgroup=False
            )
        success("Added template group permissions.")

    render_result(
        AddUsergroupPermissionsResult(
            usergroup=usergroup,
            hostgroups=hgroups,
            templategroups=tgroups,
            permission=permission,
        ),
    )


@app.command("create_usergroup", rich_help_panel=HELP_PANEL)
def create_usergroup(
    ctx: typer.Context,
    usergroup: str = typer.Argument(
        help="Name of the user group to create.",
        show_default=False,
    ),
    gui_access: GUIAccess = typer.Option(
        GUIAccess.DEFAULT.value, "--gui", help="GUI access for the group."
    ),
    disabled: bool = typer.Option(
        False,
        "--disabled",
        help="Create the user group in a disabled state.",
    ),
    # V2 legacy args
    args: Optional[list[str]] = ARGS_POSITIONAL,
) -> None:
    """Create a user group."""
    # We already have name and GUI access, so we expect 1 more arg at most
    if args:
        if len(args) != 1:
            exit_err(
                "Invalid number of positional arguments. Please use options instead."
            )
        disabled = parse_bool_arg(args[0])

    with suppress(ZabbixNotFoundError):
        app.state.client.get_usergroup(usergroup)
        exit_err(f"User group {usergroup!r} already exists.")

    with app.status("Creating user group"):
        usergroupid = app.state.client.create_usergroup(
            usergroup, gui_access=gui_access, disabled=disabled
        )
    success(f"Created user group {usergroup!r} ({usergroupid}).")


@app.command("remove_usergroup", rich_help_panel=HELP_PANEL)
def remove_usergroup(
    ctx: typer.Context,
    usergroup: str = typer.Argument(
        help="Name of the user group(s) to delete. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="Force deletion if any user groups contain users.",
    ),
) -> None:
    """Remove a user group."""
    from zabbix_cli.commands.results.usergroup import RemoveUsergroupResult
    from zabbix_cli.pyzabbix.types import Usergroup

    # Require force for * (delete all) to avoid accidental mass deletions
    if usergroup.strip() == "*":
        if not force:
            exit_err(
                "The [option]--force[/option] option is required when deleting all user groups."
            )
        # ALSO require confirmation
        from zabbix_cli.output.prompts import bool_prompt

        if not bool_prompt(
            "Are you sure you want to delete ALL user groups? This action cannot be undone."
        ):
            exit_err("Aborted by user.")

    ug_names = parse_list_arg(usergroup)
    if not ug_names:
        exit_err("At least one user group name must be specified.")

    usergroups: list[Usergroup] = []
    with app.status("Fetching user groups"):
        for usergroup in ug_names:
            try:
                ugs = app.state.client.get_usergroups(
                    usergroup, select_users=True, search=True
                )
            except ZabbixNotFoundError:
                exit_err(f"User group {usergroup!r} does not exist.")
            else:
                if not ugs:
                    exit_err(f"No user groups found matching {usergroup!r}.")
                usergroups.extend(ugs)

    # dedup
    # TODO: add dedup function with HasName interface
    seen: set[str] = set()
    unique_usergroups: list[Usergroup] = []
    for ug in usergroups:
        if ug.name not in seen:
            seen.add(ug.name)
            unique_usergroups.append(ug)
    usergroups = unique_usergroups

    # Check for users in user groups
    has_users: list[Usergroup] = []
    for ug in usergroups:
        if ug.users:
            has_users.append(ug)

    # Format names + pluralization
    ug_names_str = ", ".join(f"{ug.name!r}" for ug in has_users)
    ugs_fmt = p("user group", len(usergroups), with_count=False)
    if has_users and not force:
        exit_err(
            f"{ugs_fmt.capitalize()} with users: {ug_names_str}. [option]--force[/option] is required."
        )

    with app.status("Removing user group") as status:
        for ug in usergroups:
            status.update(f"Removing user group {ug.name!r}...")
            app.state.client.delete_usergroup(ug)

    render_result(RemoveUsergroupResult.from_result(usergroups))
    success(f"Removed {len(usergroups)} {ugs_fmt}.")


@app.command("remove_user_from_usergroup", rich_help_panel=HELP_PANEL)
def remove_user_from_usergroup(
    ctx: typer.Context,
    usernames: str = typer.Argument(
        help="Usernames to remove. Comma-separated.",
        show_default=False,
    ),
    usergroups: str = typer.Argument(
        help="User groups to remove the users from. Comma-separated.",
        show_default=False,
    ),
) -> None:
    """Remove users from usergroups.

    Ignores users not in user groups. Users and groups must exist.
    """
    from zabbix_cli.commands.results.usergroup import UsergroupRemoveUsers

    # FIXME: requires support for IDs for parity with V2
    unames = parse_list_arg(usernames)
    ugroups = parse_list_arg(usergroups)

    with app.status("Removing users from user groups"):
        users = [app.state.client.get_user(u) for u in unames]
        for ugroup in ugroups:
            try:
                app.state.client.remove_usergroup_users(ugroup, users)
            except ZabbixAPIException as e:
                exit_err(f"Failed to remove users from user group {ugroup!r}: {e}")

    render_result(
        UsergroupRemoveUsers(
            usergroups=ugroups,
            users=[u.username for u in users],
        ),
    )
    success("Removed users from user groups.")


@app.command(
    "show_usergroup",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show user group 'Admins'",
            "show_usergroup Admins",
        ),
        Example(
            "Show user groups 'Admins' and 'Operators'",
            "show_usergroup Admins,Operators",
        ),
        Example(
            "Show all user groups containing 'web' sorted by ID",
            "show_usergroup '*web*' --sort id",
        ),
    ],
)
def show_usergroup(
    ctx: typer.Context,
    usergroup: str = typer.Argument(
        help="Name or ID of the user group(s) to show. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    sort: UsergroupSorting = OPTION_SORT_UGROUPS,
) -> None:
    """Show one or more user groups by name or ID."""
    _do_show_usergroups(usergroup, sort)


@app.command(
    "show_usergroups",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show all user groups",
            "show_usergroups",
        ),
        Example(
            "Show user groups 'Admins' and 'Operators'",
            "show_usergroup Admins,Operators",
        ),
        Example(
            "Show all user groups containing 'web' sorted by ID",
            "show_usergroup '*web*' --sort id",
        ),
    ],
)
def show_usergroups(
    ctx: typer.Context,
    usergroup: Optional[str] = typer.Argument(
        None,
        help="Name or ID of the user group(s) to show. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    sort: UsergroupSorting = OPTION_SORT_UGROUPS,
    limit: Optional[int] = OPTION_LIMIT,
) -> None:
    """Show all suser groups.

    Can be filtered by name or ID.
    """
    _do_show_usergroups(usergroup, sort=sort, limit=limit)


def _do_show_usergroups(
    usergroup: Optional[str],
    sort: UsergroupSorting,
    limit: Optional[int] = None,
) -> None:
    from zabbix_cli.commands.results.usergroup import ShowUsergroupResult
    from zabbix_cli.models import AggregateResult

    ugs = parse_list_arg(usergroup)
    with app.status("Fetching user groups..."):
        usergroups = app.state.client.get_usergroups(
            *ugs, select_users=True, search=True, limit=limit
        )
    res: list[ShowUsergroupResult] = []
    for ugroup in usergroups:
        res.append(ShowUsergroupResult.from_usergroup(ugroup))
    # NOTE: why client-side sorting?
    res = sort_ugroups(res, sort)
    render_result(AggregateResult(result=res))


@app.command(
    "show_usergroup_permissions",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show permissions for user group 'Admins'",
            "show_usergroup_permissions Admins",
        ),
        Example(
            "Show permissions for user groups 'Admins' and 'Operators'",
            "show_usergroup_permissions Admins,Operators",
        ),
        Example(
            "Show permissions for all user groups sorted by ID",
            "show_usergroup_permissions * --sort id",
        ),
    ],
)
def show_usergroup_permissions(
    ctx: typer.Context,
    usergroup: str = typer.Argument(
        help="Name of user group. Comma-separated. Supports wildcards.",
        show_default=False,
    ),
    sort: UsergroupSorting = OPTION_SORT_UGROUPS,
) -> None:
    """Show permissions for one or more user groups."""
    # FIXME: URGENT Does not work properly in 6.0.22

    # NOTE: this command breaks JSON output compatibility with V2
    # In V2, rights were serialized as a string in the format of "<NAME> (<RO/RW/DENY>)"
    # under the key "permissions".
    # In V3, we follow the API and serialize it as a list of dicts under the key
    # "rights" in <6.2.0 and "hostgroup_rights" and "templategroup_rights" in >=6.2.0
    from zabbix_cli.commands.results.usergroup import ShowUsergroupPermissionsResult
    from zabbix_cli.models import AggregateResult

    ugs = parse_list_arg(usergroup)
    usergroups = app.state.client.get_usergroups(*ugs, select_rights=True, search=True)

    if not usergroups:
        exit_err("No user groups found.")

    with app.status("Fetching host groups..."):
        hostgroups = app.state.client.get_hostgroups()
    if app.state.client.version.release >= (6, 2, 0):
        with app.status("Fetching template groups..."):
            templategroups = app.state.client.get_templategroups()
    else:
        templategroups = []
    res: list[ShowUsergroupPermissionsResult] = []
    for ugroup in usergroups:
        res.append(
            ShowUsergroupPermissionsResult.from_usergroup(
                ugroup, hostgroups=hostgroups, templategroups=templategroups
            )
        )
    render_result(AggregateResult(result=sort_ugroups(res, sort)))



from __future__ import annotations

import logging
import unittest

import zabbix_cli.logs

logger = logging.getLogger(__name__)


class CollectHandler(logging.Handler):
    def __init__(self):
        super().__init__(logging.NOTSET)
        self.records = []

    def emit(self, record):
        print("emit", repr(record))
        self.records.append(record)


class TestSafeFormatting(unittest.TestCase):
    def _make_log_record(self, msg, *args):
        # locals for readability
        record_logger = "example-logger-name"
        record_level = logging.ERROR
        record_pathname = __file__
        record_lineno = 1
        record_exc_info = None
        return logging.LogRecord(
            record_logger,
            record_level,
            record_pathname,
            record_lineno,
            msg,
            args,
            record_exc_info,
        )

    def _make_safe_record(self, msg, *args):
        return zabbix_cli.logs.SafeRecord(self._make_log_record(msg, *args))

    def test_safe_record(self):
        self._make_safe_record("foo")
        self.assertTrue(True)  # reached

    def test_safe_record_basic_message(self):
        message = "foo"
        record = self._make_safe_record(message)
        self.assertEqual(message, record.getMessage())

    def test_safe_record_formatted_message(self):
        expect = "foo 01 02"
        record = self._make_safe_record("foo %02d %02d", 1, 2)
        self.assertEqual(expect, record.getMessage())

    def test_safe_record_attr(self):
        msg = ("foo",)
        args = (1, 2)
        record = self._make_safe_record(msg, *args)
        self.assertEqual(msg, record.msg)
        self.assertEqual(args, record.args)

    def test_safe_record_missing_dict(self):
        fmt = "%(msg)s-%(this_probably_does_not_exist)s"
        record = self._make_safe_record("foo")
        expect = f"foo-{None}"
        result = fmt % record.__dict__
        self.assertEqual(expect, result)

    def test_safe_formatter(self):
        fmt = "%(name)s - %(something)s - %(msg)s"
        formatter = zabbix_cli.logs.SafeFormatter(fmt)
        record = self._make_log_record("foo")

        expect = fmt % {"name": record.name, "something": None, "msg": record.msg}
        self.assertEqual(expect, formatter.format(record))



from __future__ import annotations

from pathlib import Path

from zabbix_cli.dirs import DIRS
from zabbix_cli.dirs import Directory

# Directory of all docs files
DOC_DIR = Path(__file__).parent.parent

# Directory of data files for Jinja2 templates
DATA_DIR = DOC_DIR / "data"
if not DATA_DIR.exists():
    DATA_DIR.mkdir(parents=True)

# Directory of Jinja2 templates
TEMPLATES_DIR = DOC_DIR / "templates"

# Directory of generated command doc pages
COMMANDS_DIR = DOC_DIR / "commands"
if not COMMANDS_DIR.exists():
    COMMANDS_DIR.mkdir(parents=True)


def sanitize_dirname(d: Directory) -> str:
    """Sanitize directory name for use in filenames."""
    return f"{d.name.lower().replace(' ', '_')}_dir"


def add_path_placeholders(s: str) -> str:
    """Add placeholders for file paths used by the application in a string.

    Enables somewhat consistent file paths in the documentation
    regardless of the runner environment.
    """
    for directory in DIRS:
        # Naive string replacement, then clean up double slashes if any
        s = s.replace(f"{directory.path}", f"/path/to/{sanitize_dirname(directory)}")
        s = s.replace("//", "/")
    return s



"""Utility functions for working with CLI commands."""

from __future__ import annotations

from typing import TYPE_CHECKING

import typer
import typer.core

from zabbix_cli.exceptions import ZabbixCLIError

if TYPE_CHECKING:
    import click


def get_parent_ctx(
    ctx: typer.Context | click.core.Context,
) -> typer.Context | click.core.Context:
    """Get the top-level parent context of a context."""
    if ctx.parent is None:
        return ctx
    return get_parent_ctx(ctx.parent)


def get_command_help(command: typer.models.CommandInfo) -> str:
    """Get the help text of a command."""
    if command.help:
        return command.help
    if command.callback and command.callback.__doc__:
        lines = command.callback.__doc__.strip().splitlines()
        if lines:
            return lines[0]
    if command.short_help:
        return command.short_help
    return ""


def get_command_by_name(ctx: typer.Context, name: str) -> click.core.Command:
    """Get a CLI command given its name."""
    if not isinstance(ctx.command, typer.core.TyperGroup):
        # NOTE: ideally we shouldn't leak this error to the user, but
        # can this even happen? Isn't it always a command group?
        raise ZabbixCLIError(f"Command {ctx.command.name} is not a command group.")
    if not ctx.command.commands:
        raise ZabbixCLIError(f"Command group {ctx.command.name} has no commands.")
    command = ctx.command.commands.get(name)
    if not command:
        raise ZabbixCLIError(f"Command {name} not found.")
    return command



from __future__ import annotations

from collections.abc import MutableSequence
from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Optional
from typing import Union
from typing import cast

import rich.box
from packaging.version import Version
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import JsonValue
from pydantic.fields import ComputedFieldInfo
from pydantic.fields import FieldInfo
from strenum import StrEnum
from typing_extensions import TypeVar

from zabbix_cli.logs import logger
from zabbix_cli.table import get_table
from zabbix_cli.utils.rich import get_safe_renderable

if TYPE_CHECKING:
    from rich.console import RenderableType
    from rich.table import Table


class ReturnCode(StrEnum):
    DONE = "Done"
    ERROR = "Error"


ColsType = list[str]
"""A list of column headers."""

RowContent = MutableSequence["RenderableType"]
"""A list of renderables representing the content of a row."""

RowsType = MutableSequence[RowContent]
"""A list of rows, where each row is a list of strings."""

ColsRowsType = tuple[ColsType, RowsType]
"""A tuple containing a list of columns and a list of rows, where each row is a list of strings."""


class MetaKey(str, Enum):
    """Keys used in the `json_schema_extra` dict of a field to customize
    its rendering.
    """

    JOIN_CHAR = "join_char"
    HEADER = "header"


def fmt_field_name(field_name: str) -> str:
    """Formats a field name for display in a table."""
    return field_name.capitalize().replace("_", " ")


# We wrap the results of commands in a Result object,
# but ONLY if we are rendering it as JSON. This makes the logic in the
# `render` module a bit of a mess, since the function type annotations
# are all over the place.
#
# Yeah, this all passes type checking and all that, but it's very Bad (tm)
# and way more complicated than it probably has to be.


T = TypeVar("T", bound=JsonValue)


class TableRenderable(BaseModel):
    """Base model that can be rendered as a table."""

    model_config = ConfigDict(populate_by_name=True)

    __title__: Optional[str] = None
    __show_lines__: bool = True
    __box__: rich.box.Box = rich.box.ROUNDED

    zabbix_version: ClassVar[Version] = Version(
        "7.0.0"
    )  # assume latest released version
    """Zabbix API version the data stems from.
    This is a class variable that can be overridden, which causes all
    subclasses to use the new value.

    This class variable is set by `State.login` based on the connected
    Zabbix server API version. Assumes latest released version by default.
    """
    legacy_json_format: ClassVar[bool] = False
    """Use the legacy JSON format when rendered as JSON.

    This class variable is set by `State.login` based on the
    current configuration. Assumes new JSON format by default."""

    empty_ok: bool = Field(default=False, exclude=True)
    """Don't print a message if table is empty when rendered as Table."""

    def _get_extra(self, field: str, key: MetaKey, default: T) -> T:
        f = self.model_fields.get(field, None)
        if not f:
            raise ValueError(f"Field {field!r} does not exist.")
        if not f.json_schema_extra or not isinstance(f.json_schema_extra, dict):
            return default
        # NOTE: this cast isn't super type safe, but we are expected to call this
        # method with the extra key constants defined above.
        #
        # If need be, we can add some sort of model validator that ensures
        # all JSON schema extra keys have the correct type.
        # But that will only happen once we actually encounter such a bug.
        return cast(T, f.json_schema_extra.get(key, default))

    def __all_fields__(self) -> dict[str, Union[FieldInfo, ComputedFieldInfo]]:
        """Returns all fields for the model, including computed fields,
        but excluding excluded fields.
        """
        all_fields: dict[str, Union[FieldInfo, ComputedFieldInfo]] = {
            **self.model_fields,
            **self.model_computed_fields,
        }
        return {n: f for n, f in all_fields.items() if not getattr(f, "exclude", False)}

    def __cols__(self) -> ColsType:
        """Returns the columns for the table representation of the object.

        Only override if you want to customize the column headers without
        overriding the rows. Otherwise, override `__cols_rows__`.

        By default, uses the name of the fields as the column headers,
        with the first letter capitalized.
        This can be overriden with `header` in `json_schema_extra`:

        >>> class User(TableRenderable):
        ...     userid: str = Field(json_schema_extra={"header" : "User ID"})
        ...     username: str = ""
        ...
        >>> User().__cols__()
        ["User ID", "Username"]
        """
        cols: list[str] = []

        for field_name, field in self.__all_fields__().items():
            if (
                field.json_schema_extra
                and isinstance(field.json_schema_extra, dict)
                and field.json_schema_extra.get(MetaKey.HEADER, None)
            ):
                cols.append(str(field.json_schema_extra[MetaKey.HEADER]))
            else:
                cols.append(fmt_field_name(field_name))
        return cols

    def __rows__(self) -> RowsType:
        r"""Returns the rows for the table representation of the object.

        Only override if you want to customize the rows without
        overriding the columns. Otherwise, override `__cols_rows__`.

        Render types in the following way:
            - TableRenderable: render as a table
            - BaseModel: render as JSON string
            - list: render as newline delimited string
        Everything else is rendered as a string.

        Example
        -------
        >>> class User(TableRenderable):
        ...     userid: str
        ...     username: str
        ...     groups: List[str] = []
        ...
        >>> User(userid="1", username="admin", groups=["foo", "bar", "baz"]).__rows__()
        [["1", "admin", "foo\nbar\nbaz"]]
        """  # noqa: D416
        fields: dict[str, Any | str] = {
            field_name: getattr(self, field_name, "")
            for field_name in self.__all_fields__()
        }
        for field_name, value in fields.items():
            if isinstance(value, TableRenderable):
                fields[field_name] = value.as_table()
            elif isinstance(value, BaseModel):
                # Fall back to rendering as JSON string
                logger.warning(
                    "Cannot render %s as a table.",
                    value.__class__.__name__,
                    stack_info=True,  # we want to know how we got here
                )
                fields[field_name] = value.model_dump_json(indent=2)
            elif isinstance(value, list):
                value = cast(list[Any], value)
                # A list either contains TableRenderable objects or stringable objects
                if value and all(isinstance(v, TableRenderable) for v in value):
                    # TableRenderables are wrapped in an AggregateResult to render them
                    # as a single table instead of a table per item.
                    # NOTE: we assume list contains items of the same type
                    # Rendering an aggregate result with mixed types is not supported
                    # and will probably break.
                    value = cast(list[TableRenderable], value)
                    fields[field_name] = AggregateResult(result=value).as_table()
                else:
                    # Other lists are rendered as newline delimited strings.
                    # The delimiter can be modified with the `JOIN_CHAR` meta-key in
                    # the field's `json_schema_extra`.
                    join_char = self._get_extra(field_name, MetaKey.JOIN_CHAR, "\n")
                    fields[field_name] = join_char.join(str(v) for v in value)
            else:
                fields[field_name] = str(value)
        return [list(fields.values())]  # must be a list of lists

    def __cols_rows__(self) -> ColsRowsType:
        """Returns the columns and rows for the table representation of the object.

        Example:
            >>> class User(TableRenderable):
            ...     userid: str = Field(json_schema_extra={"header" : "User ID"})
            ...     username: str = ""
            ...
            >>> User(userid="1", username="admin").__cols_rows__()
            (["UserID", "Username"], [["1", "admin"]])
        """
        return self.__cols__(), self.__rows__()

    def as_table(self) -> Table:
        """Renders a Rich table given the rows and cols generated for the object."""
        cols, rows = self.__cols_rows__()
        for row in rows:
            for i, cell in enumerate(row):
                row[i] = get_safe_renderable(cell)

        return get_table(
            cols=cols,
            rows=rows,
            title=self.__title__,
            show_lines=self.__show_lines__,
            box=self.__box__,
        )

    # We should implement the rich renderable protocol...


DataT = TypeVar("DataT", default=TableRenderable)


class BaseResult(TableRenderable):
    message: str = Field(default="")
    """Field that signals that the result should be printed as a message, not a table."""
    errors: list[str] = Field(default_factory=list)
    return_code: ReturnCode = ReturnCode.DONE
    table: bool = Field(default=True, exclude=True)

    model_config = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True, extra="allow"
    )


class Result(BaseResult, Generic[DataT]):
    """A result wrapping a single data object."""

    result: Optional[Union[DataT, list[DataT]]] = None

    # https://docs.pydantic.dev/latest/concepts/serialization/#serialize_as_any-runtime-setting
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        return super().model_dump(serialize_as_any=True, **kwargs)

    def model_dump_json(self, **kwargs: Any) -> str:
        return super().model_dump_json(serialize_as_any=True, **kwargs)


TableRenderableT = TypeVar("TableRenderableT", bound=TableRenderable)


class AggregateResult(BaseResult, Generic[TableRenderableT]):
    """Resut wrapping multiple table renderables.

    Used for compatibility with the legacy JSON format,
    as well as implementing table rendering for multiple
    results.
    """

    result: list[TableRenderableT] = Field(default_factory=list)

    def __cols_rows__(self) -> ColsRowsType:
        cols: ColsType = []
        rows: RowsType = []

        for result in self.result:
            c, r = result.__cols_rows__()
            if not cols:
                cols = c
            if r:
                rows.append(r[0])  # NOTE: why not add all rows?
        return cols, rows



"""In order to mimick the API of Zabbix-cli < 3.0.0, we define a single
app object here, which we share between the different command modules.

Thus, every command is part of the same command group.
"""

from __future__ import annotations

import inspect
import logging
from collections.abc import Iterable
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Protocol
from typing import Union

import typer
from typer.core import TyperCommand
from typer.core import TyperGroup
from typer.main import Typer
from typer.main import get_group
from typer.models import CommandFunctionType
from typer.models import CommandInfo as TyperCommandInfo
from typer.models import Default

from zabbix_cli.app.plugins import PluginLoader
from zabbix_cli.logs import logger
from zabbix_cli.state import State
from zabbix_cli.state import get_state

if TYPE_CHECKING:
    from rich.console import RenderableType
    from rich.status import Status
    from rich.style import StyleType

    from zabbix_cli.config.model import Config
    from zabbix_cli.config.model import PluginConfig


class Example(NamedTuple):
    """Example command usage."""

    description: str
    command: str
    return_value: Optional[str] = None

    def __str__(self) -> str:
        return f"  [i]{self.description}[/]\n\n    [example]{self.command}[/]"


# TODO: Trigger example rendering only when user calls --help, so we don't build
#       the help text for every command on startup.
#       Need to investigate ctx.get_help() and how it's used.
#       The question is whether we have to monkeypatch this or if we can do it with
#       the current typer/click API
class CommandInfo(TyperCommandInfo):
    def __init__(
        self, *args: Any, examples: Optional[list[Example]] = None, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.examples = examples or []
        self.set_command_help()

    def set_command_help(self) -> None:
        if not self.help:
            self.help = inspect.getdoc(self.callback) or ""
        if not self.short_help:
            self.short_help = self.help.split("\n")[0]
        self._set_command_examples()

    def _set_command_examples(self) -> None:
        if not self.examples or not self.help:
            return
        examples = [str(e) for e in self.examples]
        examples.insert(0, "\n\n[bold underline]Examples[/]")

        self.help = self.help.strip()
        self.help += "\n\n".join(examples)


class StatusCallable(Protocol):
    """Function that returns a Status object.

    Protocol for rich.console.Console.status method.
    """

    def __call__(
        self,
        status: RenderableType,
        *,
        spinner: str = "dots",
        spinner_style: StyleType = "status.spinner",
        speed: float = 1.0,
        refresh_per_second: float = 12.5,
    ) -> Status: ...


class StatefulApp(typer.Typer):
    """A Typer app that provides access to the global state."""

    parent: Optional[StatefulApp]
    plugins: dict[str, ModuleType]

    # NOTE: might be a good idea to add a typing.Unpack definition for the kwargs?
    def __init__(self, **kwargs: Any) -> None:
        self.parent = None
        self._plugin_loader = PluginLoader()
        super().__init__(**kwargs)

    @property
    def logger(self) -> logging.Logger:
        return logger

    # Methods for adding subcommands and keeping track of hierarchy
    def add_typer(self, typer_instance: Typer, **kwargs: Any) -> None:
        kwargs.setdefault("no_args_is_help", True)
        if isinstance(typer_instance, StatefulApp):
            typer_instance.parent = self
        return super().add_typer(typer_instance, **kwargs)

    def add_subcommand(self, app: typer.Typer, *args: Any, **kwargs: Any) -> None:
        kwargs.setdefault("rich_help_panel", "Subcommands")
        self.add_typer(app, **kwargs)

    def load_plugins(self, config: Config) -> None:
        """Load plugins."""
        self._plugin_loader.load(config)

    def configure_plugins(self, config: Config) -> None:
        """Configure plugins."""
        self._plugin_loader.configure_plugins(config)

    def parents(self) -> Iterable[StatefulApp]:
        """Get all parent apps."""
        app = self
        while app.parent:
            yield app.parent
            app = app.parent

    def find_root(self) -> StatefulApp:
        """Get the root app."""
        app = self
        for parent in self.parents():
            app = parent
        return app

    def as_click_group(self) -> TyperGroup:
        """Return the Typer app as a Click group."""
        return get_group(self)

    def command(
        self,
        name: Optional[str] = None,
        *,
        cls: Optional[type[TyperCommand]] = None,
        context_settings: Optional[dict[Any, Any]] = None,
        help: Optional[str] = None,
        epilog: Optional[str] = None,
        short_help: Optional[str] = None,
        options_metavar: str = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        # Rich settings
        rich_help_panel: Union[str, None] = Default(None),
        # Zabbix-cli kwargs
        examples: Optional[list[Example]] = None,
    ) -> Callable[[CommandFunctionType], CommandFunctionType]:
        if cls is None:
            cls = TyperCommand

        def decorator(f: CommandFunctionType) -> CommandFunctionType:
            self.registered_commands.append(
                CommandInfo(
                    name=name,
                    cls=cls,
                    context_settings=context_settings,
                    callback=f,
                    help=help,
                    epilog=epilog,
                    short_help=short_help,
                    options_metavar=options_metavar,
                    add_help_option=add_help_option,
                    no_args_is_help=no_args_is_help,
                    hidden=hidden,
                    deprecated=deprecated,
                    # Rich settings
                    rich_help_panel=rich_help_panel,
                    # Zabbix-cli kwargs
                    examples=examples,
                )
            )
            return f

        return decorator

    @property
    def state(self) -> State:
        return get_state()

    @property
    def api_version(self) -> tuple[int, ...]:
        """Get the current API version. Will fail if not connected to the API."""
        return self.state.client.version.release

    @property
    def status(self) -> StatusCallable:
        return self.state.err_console.status

    def get_plugin_config(self, name: str) -> PluginConfig:
        """Get a plugin's configuration by name.

        Returns an empty PluginConfig object if no config is found.
        """
        conf = self.state.config.plugins.get(name)
        if not conf:
            # NOTE: can we import this top-level? We have probably already imported
            # the config at this point? Unless we refactor config loading _again_...?
            from zabbix_cli.config.model import PluginConfig

            logger.error(f"Plugin '{name}' not found in configuration")
            return PluginConfig()
        return conf



from __future__ import annotations

from functools import cache
from typing import Any
from typing import Optional
from typing import Union
from typing import cast

import click
import typer
from pydantic import BaseModel
from pydantic import Field
from pydantic import ValidationError
from pydantic import computed_field
from pydantic import model_validator
from typer.core import TyperArgument
from typer.core import TyperCommand
from typer.core import TyperGroup
from typer.models import DefaultPlaceholder
from zabbix_cli.exceptions import ZabbixCLIError

from .markup import markup_as_plain_text
from .markup import markup_to_markdown


def get(param: Any, attr: str) -> Any:
    """Getattr that defaults to None"""
    return getattr(param, attr, None)


class ParamSummary(BaseModel):
    """Serializable representation of a click.Parameter."""

    allow_from_autoenv: Optional[bool] = None
    confirmation_prompt: Optional[bool] = None
    choices: Optional[list[str]] = None
    count: Optional[bool] = None
    default: Optional[Any] = None
    envvar: Optional[str]
    expose_value: bool
    flag_value: Optional[Any] = None
    help: str
    hidden: Optional[bool] = None
    human_readable_name: str
    is_argument: bool
    is_eager: bool = False
    is_bool_flag: Optional[bool] = None
    is_option: Optional[bool]
    max: Optional[int] = None
    min: Optional[int] = None
    metavar: Optional[str]
    multiple: bool
    name: Optional[str]
    nargs: int
    opts: list[str]
    prompt: Optional[str] = None
    prompt_required: Optional[bool] = None
    required: bool
    secondary_opts: list[str] = []
    show_choices: Optional[bool] = None
    show_default: Optional[bool] = None
    show_envvar: Optional[bool] = None
    type: str

    @classmethod
    def from_param(cls, param: click.Parameter) -> ParamSummary:
        """Construct a new ParamSummary from a click.Parameter."""
        try:
            help_ = param.help or ""  # type: ignore
        except AttributeError:
            help_ = ""

        is_argument = isinstance(param, (click.Argument, TyperArgument))
        return cls(
            allow_from_autoenv=get(param, "allow_from_autoenv"),
            confirmation_prompt=get(param, "confirmation_prompt"),
            count=get(param, "count"),
            choices=get(param.type, "choices"),
            default=param.default,
            envvar=param.envvar,  # TODO: support list of envvars
            expose_value=param.expose_value,
            flag_value=get(param, "flag_value"),
            help=help_,
            hidden=get(param, "hidden"),
            human_readable_name=param.human_readable_name,
            is_argument=is_argument,
            is_bool_flag=get(param, "is_bool_flag"),
            is_eager=param.is_eager,
            is_option=get(param, "is_option"),
            max=get(param.type, "max"),
            min=get(param.type, "min"),
            metavar=param.metavar,
            multiple=param.multiple,
            name=param.name,
            nargs=param.nargs,
            opts=param.opts,
            prompt=get(param, "prompt"),
            prompt_required=get(param, "prompt_required"),
            required=param.required,
            secondary_opts=param.secondary_opts,
            show_choices=get(param, "show_choices"),
            show_default=get(param, "show_default"),
            show_envvar=get(param, "show_envvar"),
            type=param.type.name,
        )

    @property
    def help_plain(self) -> str:
        return markup_as_plain_text(self.help)

    @property
    def help_md(self) -> str:
        return markup_to_markdown(self.help)

    @model_validator(mode="before")
    @classmethod
    def _fmt_metavar(cls, data: Any) -> Any:
        if isinstance(data, dict):
            metavar = data.get("metavar", "") or data.get(  # pyright: ignore[reportUnknownMemberType,reportUnknownVariableType]
                "human_readable_name", ""
            )
            assert isinstance(metavar, str), "metavar must be a string"
            metavar = metavar.upper()
            if data.get("multiple"):  # pyright: ignore[reportUnknownMemberType]
                new_metavar = f"<{metavar},[{metavar}...]>"
            else:
                new_metavar = f"<{metavar}>"
            data["metavar"] = new_metavar
        return data  # pyright: ignore[reportUnknownVariableType]

    @computed_field
    @property
    def show(self) -> bool:
        if self.hidden:
            return False
        if "deprecated" in self.help.lower():
            return False
        return True


# TODO: split up CommandSummary into CommandSummary and CommandSearchResult
# so that the latter can have the score field
class CommandSummary(BaseModel):
    """Convenience class for accessing information about a command."""

    category: Optional[str] = None  # not part of TyperCommand
    deprecated: bool
    epilog: Optional[str]
    help: str
    hidden: bool
    name: str
    options_metavar: str
    params: list[ParamSummary] = Field([], exclude=True)
    score: int = 0  # match score (not part of TyperCommand)
    short_help: Optional[str]

    @model_validator(mode="before")
    @classmethod
    def _replace_placeholders(cls, values: Any) -> Any:
        """Replace DefaultPlaceholder values with empty strings."""
        if not isinstance(values, dict):
            return values
        values = cast(dict[str, Any], values)
        for key, value in values.items():
            if isinstance(value, DefaultPlaceholder):
                # Use its value, otherwise empty string
                values[key] = value.value or ""
        return values

    @classmethod
    def from_command(
        cls, command: TyperCommand, name: str | None = None, category: str | None = None
    ) -> CommandSummary:
        """Construct a new CommandSummary from a TyperCommand."""
        try:
            return cls(
                category=category,
                deprecated=command.deprecated,
                epilog=command.epilog or "",
                help=command.help or "",
                hidden=command.hidden,
                name=name or command.name or "",
                options_metavar=command.options_metavar or "",
                params=[ParamSummary.from_param(p) for p in command.params],
                short_help=command.short_help or "",
            )
        except ValidationError as e:
            raise ZabbixCLIError(
                f"Failed to construct command summary for {name or command.name}: {e}"
            ) from e

    @property
    def help_plain(self) -> str:
        return markup_as_plain_text(self.help)

    @property
    def help_md(self) -> str:
        return markup_to_markdown(self.help)

    @computed_field
    @property
    def usage(self) -> str:
        parts = [self.name]

        # Assume arg list is sorted by required/optional
        # `<POSITIONAL_ARG1> <POSITIONAL_ARG2> [OPTIONAL_ARG1] [OPTIONAL_ARG2]`
        for arg in self.arguments:
            metavar = arg.metavar or arg.human_readable_name
            parts.append(metavar)

        # Command with both required and optional options:
        # `--option1 <opt1> --option2 <opt2> [OPTIONS]`
        has_optional = False
        for option in self.options:
            if option.required:
                metavar = option.metavar or option.human_readable_name
                if option.opts:
                    s = f"{max(option.opts)} {metavar}"
                else:
                    # this shouldn't happen, but just in case. A required
                    # option without any opts is not very useful.
                    # NOTE: could raise exception here instead
                    s = metavar
                parts.append(s)
            else:
                has_optional = True
        if has_optional:
            parts.append("[OPTIONS]")

        return " ".join(parts)

    @computed_field
    @property
    def options(self) -> list[ParamSummary]:
        return [p for p in self.params if _include_opt(p)]

    @computed_field
    @property
    def arguments(self) -> list[ParamSummary]:
        return [p for p in self.params if _include_arg(p)]


def _include_arg(arg: ParamSummary) -> bool:
    """Determine if an argument or option should be included in the help output."""
    if not arg.is_argument:
        return False
    return arg.show


def _include_opt(opt: ParamSummary) -> bool:
    """Determine if an argument or option should be included in the help output."""
    if opt.is_argument:
        return False
    return opt.show


def get_parent_ctx(
    ctx: typer.Context | click.core.Context,
) -> typer.Context | click.core.Context:
    """Get the top-level parent context of a context."""
    if ctx.parent is None:
        return ctx
    return get_parent_ctx(ctx.parent)


def get_command_help(command: typer.models.CommandInfo) -> str:
    """Get the help text of a command."""
    if command.help:
        return command.help
    if command.callback and command.callback.__doc__:
        lines = command.callback.__doc__.strip().splitlines()
        if lines:
            return lines[0]
    if command.short_help:
        return command.short_help
    return ""


@cache
def get_app_commands(app: typer.Typer) -> list[CommandSummary]:
    """Get a list of commands from a typer app."""
    return _get_app_commands(app)


def _get_app_commands(
    app: typer.Typer,
    cmds: list[CommandSummary] | None = None,
) -> list[CommandSummary]:
    if cmds is None:
        cmds = []

    # NOTE: incorrect type annotation for get_command() here:
    # The function can return either a TyperGroup or click.Command
    cmd = typer.main.get_command(app)
    cmd = cast(Union[TyperGroup, click.Command], cmd)

    groups: dict[str, TyperCommand] = {}
    try:
        groups = cmd.commands  # type: ignore
    except AttributeError:
        pass

    # If we have subcommands, we need to go deeper.
    for command in groups.values():
        if command.deprecated:  # skip deprecated commands
            continue
        category = command.rich_help_panel

        # rich_help_panel can also be a DefaultPlaceholder
        # even if the type annotation says it's str | None
        if category and not isinstance(category, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise ValueError(f"{command.name} is missing a rich_help_panel (category)")

        cmds.append(
            CommandSummary.from_command(command, name=command.name, category=category)
        )

    return sorted(cmds, key=lambda x: x.name)


def get_app_callback_options(app: typer.Typer) -> list[typer.models.OptionInfo]:
    """Get the options of the main callback of a Typer app."""
    options: list[typer.models.OptionInfo] = []

    if not app.registered_callback:
        return options

    callback = app.registered_callback.callback

    if not callback:
        return options
    if not hasattr(callback, "__defaults__") or not callback.__defaults__:
        return options

    for option in callback.__defaults__:
        options.append(option)
    return options



from __future__ import annotations

from inline_snapshot import snapshot
from pytest import LogCaptureFixture
from zabbix_cli.app.app import StatefulApp
from zabbix_cli.config.model import PluginConfig
from zabbix_cli.state import State


def test_get_plugin_config(
    app: StatefulApp, state: State, caplog: LogCaptureFixture
) -> None:
    """Test that we can get a plugin's configuration."""
    # Add a plugin configuration
    state.config.plugins.root = {
        # From module specification
        "my_commands": PluginConfig(
            module="path.to.my_commands",
        ),
        # From path
        "my_commands2": PluginConfig(
            module="path/to/my_commands2.py",
        ),
        # From package with entrypoint
        "my_commands3": PluginConfig(),
    }

    # With name
    config = app.get_plugin_config("my_commands")
    assert config.module == "path.to.my_commands"

    # Missing config returns empty config
    config = app.get_plugin_config("missing")
    assert config.module == ""
    assert caplog.records[-1].message == snapshot(
        "Plugin 'missing' not found in configuration"
    )



from __future__ import annotations

from zabbix_cli.commands import bootstrap_commands  # type: ignore # noqa: E402, F401

from .app import *  # noqa: F403 # wildcard import to avoid circular import (why?)
from .app import StatefulApp  # explicit import for type checker

app = StatefulApp(
    name="zabbix-cli",
    help="Zabbix-CLI is a command line interface for Zabbix.",
    add_completion=True,
    rich_markup_mode="rich",
)

# Import commands to register them with the app
from zabbix_cli.commands import cli  # type: ignore # noqa: E402, F401, I001
from zabbix_cli.commands import export  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import host  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import host_interface  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import hostgroup  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import host_monitoring  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import item  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import macro  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import maintenance  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import media  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import problem  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import proxy  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import template  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import templategroup  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import user  # type: ignore # noqa: E402, F401
from zabbix_cli.commands import usergroup  # type: ignore # noqa: E402, F401


# Import dev commands
# TODO: Disable by default, enable with a flag.
bootstrap_commands()



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Optional


class BasePatcher(ABC):
    """Context manager that logs and prints diagnostic info if an exception
    occurs.
    """

    def __init__(self, description: str) -> None:
        self.description = description

    @abstractmethod
    def __package_info__(self) -> str:
        raise NotImplementedError

    def __enter__(self) -> BasePatcher:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        if not exc_type:
            return True
        import sys

        import rich
        from rich.table import Table

        from zabbix_cli.__about__ import __version__

        # Rudimentary, but provides enough info to debug and fix the issue
        console = rich.console.Console(stderr=True)
        console.print_exception()
        console.print()
        table = Table(
            title="Diagnostics",
            show_header=False,
            show_lines=False,
        )
        table.add_row(
            "[b]Package [/]",
            self.__package_info__(),
        )
        table.add_row(
            "[b]zabbix-cli [/]",
            __version__,
        )
        table.add_row(
            "[b]Python [/]",
            sys.version,
        )
        table.add_row(
            "[b]Platform [/]",
            sys.platform,
        )
        console.print(table)
        console.print(f"[bold red]ERROR: Failed to patch {self.description}[/]")
        raise SystemExit(1)


def get_patcher(info: str) -> type[BasePatcher]:
    """Returns a patcher for a given package."""

    class Patcher(BasePatcher):
        def __package_info__(self) -> str:
            return info

    return Patcher



"""DEPRECATED: Set up Zabbix-CLI configuration."""

from __future__ import annotations

import subprocess
from pathlib import Path
from typing import Optional

import typer

from zabbix_cli.output.console import warning

app = typer.Typer(name="zabbix-cli-init", help="Set up Zabbix-CLI configuration")


@app.callback(invoke_without_command=True)
def main_callback(
    zabbix_url: Optional[str] = typer.Option(
        None,
        "--url",
        "--zabbix-url",
        "-z",
        help="Zabbix API URL to use",
    ),
    zabbix_user: Optional[str] = typer.Option(
        None,
        "--user",
        "--zabbix-user",
        help="Zabbix API username to use",
    ),
    config_file: Optional[Path] = typer.Option(
        None,
        "--config-file",
        "-c",
        help="Use non-default configuration file location",
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        "-o",
        help="Overwrite existing configuration file",
    ),
    wizard: bool = typer.Option(
        False,
        "--wizard/--no-wizard",
        help="Run the interactive wizard to set up the configuration file",
    ),
) -> None:
    warning(
        "[command]zabbix-cli-init[/] is deprecated. Use [command]zabbix-cli init[/]."
    )

    # HACK: run the CLI with the init command
    args = ["zabbix-cli", "init"]
    if zabbix_url:
        args.extend(["--url", zabbix_url])
    if zabbix_user:
        args.extend(["--user", zabbix_user])
    if config_file:
        args.extend(["--config-file", str(config_file)])
    if overwrite:
        args.append("--overwrite")
    if not wizard:
        args.append("--no-wizard")

    subprocess.run(args)


def main() -> None:
    """Run the CLI."""
    app()


if __name__ == "__main__":
    main()



"""Control the formatting of console output."""

from __future__ import annotations

from pathlib import Path


def path_link(path: Path, *, absolute: bool = True) -> str:
    """Return a link to a path."""
    abspath = path.resolve().absolute()
    if absolute:
        path_str = str(abspath)
    else:
        path_str = str(path)
    return f"[link=file://{abspath}]{path_str}[/link]"
